# encoding: utf-8
# Copyright (C) 2016 John Törnblom
'''
Keep version info in one single place
'''


name = 'pyrsl'
date = '2018-07-30'
version = '2.1'
release = '2.1.0' #  # ensure that this is the same as in setup.py

complete_string = '%s v%s (%s)' % (name, release, date)
